import os
import subprocess
import glob
from astropy.io import fits
from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.table import Table
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colormaps
from matplotlib.colors import SymLogNorm
from matplotlib.patches import Rectangle
import random as rd
from sklearn.cluster import DBSCAN
from sklearn.linear_model import RANSACRegressor

# To work, this file requires you to have SExtractor installed on your machine.

def run_sextractor(detect_img      : str, weight_img   : str, 
                   output_cat      : str, 
                   config_folder   : str,
                   dir_chckimg     : str,
                   detect_thresh   : float = 12.0,
                   analysis_thresh : float = 2.5,
                   verbose         : bool = False) -> None :
    """
    Run SExtractor to create a FITS_LDAC catalog, with vignets to create a PSF.

    detect_img : filename of the detection image to use
    weight_img : filename of the weight map
    output_cat : filename of the output catalog
    config_folder : path to the SEtractor config filed folder
    dir_chckimg : path to folder to save checkimages (will be created if non-existing)
    detect_thresh : value of the DETECT_THRESH to use
    analysis_thresh : value of the ANALYSIS_THRESH to use
    verbose : verbose parameter
    """
    # Find SExtractor installed version
    cmd = ""
    if subprocess.run('sex --version', shell=True, capture_output=True).returncode==0: cmd = "sex"
    elif subprocess.run('source-extractor --version', shell=True, capture_output=True).returncode==0: cmd = "source-extractor"
    else:
        print("SExtractor not found ! Make sure it is installed.")
        return None
    os.makedirs("/".join(output_cat.split("/")[:-1]), exist_ok=True)
    os.makedirs(dir_chckimg, exist_ok=True)
    study_name = '.'.join(detect_img.split('/')[-1].split('.')[:-1])
    seg_img = f'{dir_chckimg}/{study_name}_seg.fits'
    verbose_type = 'NORMAL' if verbose else 'QUIET'
    os.environ['LD_LIBRARY_PATH'] = '/usr/lib' # Link 'libcfitsio.so.10'
    subprocess.run(f'{cmd} {detect_img} -c {config_folder}/PSFEx-Cat-JWST.sex \
                   -CATALOG_NAME {output_cat} \
                   -WEIGHT_IMAGE {weight_img} -WEIGHT_TYPE MAP_WEIGHT \
                   -PARAMETERS_NAME {config_folder}/PSFEx-Cat-JWST-SW.param \
                   -FILTER_NAME {config_folder}/gauss_1.5_3x3.conv \
                   -STARNNW_NAME {config_folder}/default.nnw \
                   -CHECKIMAGE_NAME {seg_img} \
                   -DETECT_THRESH {detect_thresh} \
                   -ANALYSIS_THRESH {analysis_thresh} \
                   -VERBOSE_TYPE {verbose_type}',
                   shell=True)
    return output_cat
    
def show_vignets(data, grid_n):
    """
    Display random vignets from a SExtractor catalog

    data : data contained in the SExtractor catalog.
           Can be created with astropy.io.fits.open and by selecting hdu.data
    grid_n : tuple (a,b) defining the number of rows and columns of the vignets grid
    """
    n = len(data)
    a, b = grid_n
    selected = rd.sample(range(n), a*b)
    fig, ax = plt.subplots(a,b,figsize=(4*b,4*a))
    for i in range(a):
        for j in range(b):
            ax[i,j].imshow(data['VIGNET'][selected[i*b+j]], 
                           origin='lower', cmap=colormaps['bone_r'], 
                           norm=SymLogNorm(linthresh=1e-2, linscale=0.1, vmin=-0.1, vmax=0.5))
            ax[i,j].set_title(selected[i*b+j])
    plt.show()

def plot_MuvMAG(data, star_selections=[], mag_bounds=(17,30), mu_bounds=(13,25), ax_custom=None):
    """
    Plots MU_MAX v. MAG_AUTO, with selections overlayed
    MU_MAX = max pixel value in mag/arcsec² (MAG_ZEROPOINT - 2.5*log10(FLUX_MAX*PIXELSCALE²))
    MAG_AUTO = total flux in mag (MAG_ZEROPOINT - 2.5*log10(FLUX_AUTO))

    data : data contained in the SExtractor catalog.
           Can be created with astropy.io.fits.open and by selecting hdu.data
    star_selections : list of dictionary of the form below to define selections
        {<name> : {'label':<label>, 'color':<color>, 'flag':<array_of_indices>}}
    mag_bounds : bounds for MAG_AUTO for the plot
    mu_bounds : bounds for MU_MAX for the plot
    ax_custom : matplotlib.pyplot.Axes to display the plot.
                If ommited, the function will create a new plot
    """
    ax = plt.subplots(figsize=(8,6))[1] if ax_custom is None else ax_custom
    ax.hexbin(data['MAG_AUTO'], data['MU_MAX'], bins='log', mincnt=1, extent=(mag_bounds[0],mag_bounds[1],mu_bounds[0],mu_bounds[1]), cmap='inferno', lw=0.01)
    # ax.plot(data['MAG_AUTO'], data['MU_MAX'], marker='o', ms=0.5, ls="", alpha=0.5, c='k', label=f'All [{len(data)}]', rasterized=True)
    for name in star_selections:
        selection = star_selections[name]
        ax.plot(data[selection['flag']]['MAG_AUTO'], data[selection['flag']]['MU_MAX'], 
                   marker='o', ms=1.0, ls="", c=selection['color'], label=f"{selection['label']} [{len(selection['flag'])}]",
                   rasterized=True)
    ax.set_xlabel('MAG_AUTO')
    ax.set_ylabel('MU_MAX')
    ax.set_xlim(mag_bounds[0],mag_bounds[1])
    ax.set_ylim(mu_bounds[0],mu_bounds[1])
    ax.legend(loc='best')
    if ax_custom is None: plt.show()

def hist_CLASS_STAR(data, star_selections=[], hist_bound=500, ax_custom=None):
    """
    Plot a histogram of CLASS_STAR, with selections overlayed

    data : data contained in the SExtractor catalog.
           Can be created with astropy.io.fits.open and by selecting hdu.data
    star_selections : list of dictionary of the form below to define selections
        {<name> : {'label':<label>, 'color':<color>, 'flag':<array_of_indices>}}
    hist_bounds : bounds for the number of counts
    ax_custom : matplotlib.pyplot.Axes to display the plot.
                If ommited, the function will create a new plot
    """
    # star_selection = {<name> : {'label':<label>, 'color':<color>, 'flag':<array_of_indices>}}
    ax = plt.subplots(figsize=(8,6))[1] if ax_custom is None else ax_custom
    ax.hist(data['CLASS_STAR'], bins=100, color='k', label=f'All [{len(data)}]')
    for name in star_selections:
        selection = star_selections[name]
        ax.hist(data[selection['flag']]['CLASS_STAR'], bins=100, range=(0,1),
                color=selection['color'], label=f"{selection['label']} [{len(selection['flag'])}]")
    ax.set_ylabel('CLASS_STAR')
    ax.set_ylim(0,hist_bound)
    ax.legend(loc='best')
    if ax_custom is None: plt.show()

def plot_SNR_radius(data, star_selections=[], rad_bound=20, ax_custom=None):
    """
    Plots SNR_WIN v. FLUX_RADIUS, with selections overlayed

    data : data contained in the SExtractor catalog.
           Can be created with astropy.io.fits.open and by selecting hdu.data
    star_selections : list of dictionary of the form below to define selections
        {<name> : {'label':<label>, 'color':<color>, 'flag':<array_of_indices>}}
    rad_bound : maximum value for FLUX_RADIUS
    ax_custom : matplotlib.pyplot.Axes to display the plot.
                If ommited, the function will create a new plot
    """
    ax = plt.subplots(figsize=(8,6))[1] if ax_custom is None else ax_custom
    ax.hexbin(data['FLUX_RADIUS'], data['SNR_WIN'], bins='log', mincnt=1, yscale='log', extent=(0,rad_bound,0,6), cmap='inferno', lw=0.01)
    # ax.plot(data['FLUX_RADIUS'], data['SNR_WIN'], marker='o', ms=0.5, ls="", alpha=0.5, c='k', label=f'All [{len(data)}]', rasterized=True)
    for name in star_selections:
        selection = star_selections[name]
        ax.plot(data[selection['flag']]['FLUX_RADIUS'], data[selection['flag']]['SNR_WIN'],
                   marker='o', ms=1.0, ls="", c=selection['color'], label=f"{selection['label']} [{len(selection['flag'])}]",
                   rasterized=True)
    ax.set_xlabel('FLUX_RADIUS')
    ax.set_ylabel('SNR_WIN')
    ax.set_xlim(0, rad_bound)
    ax.set_yscale('log')
    ax.legend()
    if ax_custom is None: plt.show()

def plot_MuvMAG_manual(data, mag_bounds=(19,28), 
                       ax_custom=None, plot_mag_bounds=(18,30), plot_y_bounds=(-8,1)):
    """ 
    Plots MU_MAX-MAG_AUTO v. MAG_AUTO with guides to manually select the star line

    data : data contained in the SExtractor catalog.
           Can be created with astropy.io.fits.open and by selecting hdu.data
    mag_bounds : bounds for MAG_AUTO for the point-like sources selection
    plot_mag_bounds : bounds for MAG_AUTO for the plot
    plot_y_bounds : bounds for (MU_MAX-MAG_AUTO) for the plot
    ax_custom : matplotlib.pyplot.Axes to display the plot.
                If ommited, the function will create a new plot
    """
    ax = plt.subplots(figsize=(8,6))[1] if ax_custom is None else ax_custom
    ax.hexbin(data['MAG_AUTO'], data['MU_MAX'], bins='log', mincnt=1, extent=(plot_mag_bounds[0],plot_mag_bounds[1],plot_y_bounds[0],plot_y_bounds[1]), cmap='inferno', lw=0.01)
    # ax.plot(data['MAG_AUTO'], data['MU_MAX']-data['MAG_AUTO'], marker='o', ls="", ms=0.5, c='k', rasterized=True)
    ax.axvline(mag_bounds[0], color='r', linewidth=0.5)
    ax.axvline(mag_bounds[1], color='r', linewidth=0.5)
    ax.set_xlabel('MAG_AUTO')
    ax.set_ylabel('MU_MAX - MAG_AUTO')
    ax.set_xlim(plot_mag_bounds[0], plot_mag_bounds[1])
    ax.set_ylim(plot_y_bounds[0], plot_y_bounds[1])
    ax.grid(axis='y', color='r', linewidth=0.5, linestyle=(0, (5, 5)), alpha=1)
    ax.set_axisbelow(True)
    if ax_custom is None: plt.show()

def find_star_line(data, eps_DBSCAN=0.1, y_max=-4, mag_fit=25, verbose=False, 
                   plot=False, plot_mag_bounds=(18,30), plot_y_bounds=(-8,1),
                   save=False, save_name=None):
    """
    Finds the star line on the MU_MAX v. MAG_AUTO plot.
    1. Remove the main galaxy cluster using DBSCAN (eps_DBSCAN parameter).
    2. RANSAC regression to find the star line on the remaining points,
        and with MU_MAX-MAG_AUTO < y_max
    3. Calculates the star line position at MAG_AUTO = mag_fit

    data : data contained in the SExtractor catalog.
           Can be created with astropy.io.fits.open and by selecting hdu.data
    eps_DBSCAN : epsilon parameter for DBSCAN (the higher, the bigger the selected cluster is)
    y_max : MU_MAX-MAG_AUTO threshold for linear regression
    mag_fit : value of MAG_AUTO to calculate the star line position after linear regression
    verbose : verbose parameter
    plot : parameter to display plot to check if everything worked well
    plot_mag_bounds : bounds for MAG_AUTO for the plot
    plot_y_bounds : bounds for (MU_MAX-MAG_AUTO) for the plot
    save : parameter to save the plot to check if everything worked well
    save_name : filename for the saved plot. Mandatory if save==True

    Returns : value of the star line (MU_MAX - MAG_AUTO)
    """
    # Remove galaxy cluster
    values = np.vstack([data['MAG_AUTO'], data['MU_MAX']-data['MAG_AUTO']])
    clustering = DBSCAN(eps=eps_DBSCAN).fit(values.T)
    lab, count = np.unique(clustering.labels_, return_counts=True)
    if verbose : print(f'DBSCAN clustering : {lab, count}')
    # Find star line
    cloud_idx = np.argmax(count[lab!=-1])
    A = data[clustering.labels_ != cloud_idx]
    B = A[A['MU_MAX']-A['MAG_AUTO'] < y_max]
    if plot | save:
        fig, ax = plt.subplots(figsize=(8,6))
        ax.plot(A['MAG_AUTO'], A['MU_MAX']-A['MAG_AUTO'], ms=0.5, ls="", marker='o', c='k', rasterized=True)
        ax.axhline(y_max, color='r', linewidth=0.5, linestyle='--')
        ax.set_xlim(plot_mag_bounds[0], plot_mag_bounds[1])
        ax.set_ylim(plot_y_bounds[0], plot_y_bounds[1])
        ax.set_xlabel('MAG_AUTO')
        ax.set_ylabel('MU_MAX - MAG_AUTO')
        ax.set_title("Star line detection after galaxy cluster removal")
        if save : fig.savefig(save_name, bbox_inches='tight', dpi=100)
        if plot : plt.show()
    ransac = RANSACRegressor()
    ransac.fit(B['MAG_AUTO'].reshape(-1,1), (B['MU_MAX']-B['MAG_AUTO']).reshape(-1,1))
    if (slope:=abs(ransac.estimator_.coef_[0][0])) > 0.1: 
        print(f"\33[33m WARNING \33[0m : Star line may be wrong (Slope = {slope:.3f})")
    c = ransac.predict([[mag_fit]])[0,0]
    if verbose: print(f"{'RANSAC Slope':<13}: {ransac.estimator_.coef_[0][0]:.3f}")
    if verbose: print(f"{'Star line':<13}: {c:.3f}")
    return c

def MUvMAG_star_selection(data, star_line, y_offsets=(-0.5,0.3), mag_bounds=(19,28), snr_min=1e2, 
                          plot=False, plot_mag_bounds=(18,30), plot_y_bounds=(-8,1),
                          save=False, save_name=None):
    """
    Selects the point-like sources according to the MU_MAX v. MAG_AUTO box

    data : data contained in the SExtractor catalog.
           Can be created with astropy.io.fits.open and by selecting hdu.data
    star_line : star line value of MU_MAX - MAG_AUTO
    y_offsets : negative and positive offsets from star_line for the selection box
    mag_bounds : MAG_AUTO bounds for the selection box
    snr_min : minimum value for SNR_WIN to select the sources
    plot : parameter to display plot to check if everything worked well
    plot_mag_bounds : bounds for MAG_AUTO for the plot
    plot_y_bounds : bounds for (MU_MAX-MAG_AUTO) for the plot
    save : parameter to save the plot to check if everything worked well
    save_name : filename for the saved plot. Mandatory if save==True

    Returns : indices of the detected point-like sources in data
    """
    if plot | save:
        fig, ax = plt.subplots(figsize=(8,6))
        ax.plot(data['MAG_AUTO'], data['MU_MAX']-data['MAG_AUTO'], ms=0.5, ls="", marker='o', c='k', rasterized=True)
        ax.add_patch(Rectangle((19, star_line + y_offsets[0]), 
                                mag_bounds[1]-mag_bounds[0], 
                                y_offsets[1]-y_offsets[0], 
                                fill=False, edgecolor='r', lw=1))
        ax.set_xlabel('MAG_AUTO')
        ax.set_ylabel('MU_MAX - MAG_AUTO')
        ax.set_xlim(plot_mag_bounds[0], plot_mag_bounds[1])
        ax.set_ylim(plot_y_bounds[0], plot_y_bounds[1])
        ax.set_title("Star detection with MUvMAG")
        if save : fig.savefig(save_name, bbox_inches='tight', dpi=100)
        if plot : plt.show()
    star_MUvMAG = np.where((data['MU_MAX'] < data['MAG_AUTO'] + star_line + y_offsets[1]) & \
                           (data['MU_MAX'] > data['MAG_AUTO'] + star_line + y_offsets[0]) & \
                           (data['MAG_AUTO'] < mag_bounds[1]) & \
                           (data['MAG_AUTO'] > mag_bounds[0]) & \
                           (data['SNR_WIN'] > snr_min))[0]
    return star_MUvMAG

def save_catalog(hdul, selection, outcat_name):
    """
    Save selection as a FITS_LDAC catalog

    hdul : HDU list from the original FITS_LDAC.
           Obtained with astropy.io.fits.open
    selection : indices of the rows of the catalog to save
    outcat_name : filename of the saved catalog
    """
    hdul[2].data = hdul[2].data[selection]
    hdul[2].update_header()
    hdul.writeto(outcat_name, overwrite=True)

def extract_stars(detect_img        : str, 
                  weight_img        : str, 
                  output_cat        : str,
                  output_cat_star   : str, 
                  config_folder    : str,
                  dir_chckimg       : str,
                  detect_thresh     : float = 8.0,
                  analysis_thresh   : float = 2.5,
                  eps_DBSCAN        : float = 0.1, 
                  y_max             : float = -4.5, 
                  mag_fit           : float = 25,
                  y_offsets         : tuple = (-0.5,0.3),
                  mag_bounds        : tuple = (19,28),
                  snr_min           : float = 1e2,
                  plot_mag_bounds   : tuple = (18,30),
                  plot_mu_bounds    : tuple = (13,25), 
                  plot_y_bounds     : tuple = (-8,1),
                  plot_rad_bound    : tuple = 20,
                  save_chckimg      : bool = True,
                  plot              : bool = False,
                  clean             : bool = True,
                  run_sex           : bool = True,
                  verbose           : bool = False) -> None :
    """
    Create a point-like sources catalog from an image to be used
    in PSFEx to create the PSF of a telescope.
    
    detect_img : filename of the detection image to use
    weight_img : filename of the weight map
    output_cat : filename of the output full catalog
    output_cat_star : filename of the output point-like sources catalog
    config_folder : path to the SEtractor config filed folder
    dir_chckimg : path to folder to save checkimages (will be created if non-existing)
    detect_thresh : value of the DETECT_THRESH to use
    analysis_thresh : value of the ANALYSIS_THRESH to use
    eps_DBSCAN : epsilon parameter for DBSCAN (the higher, the bigger the selected cluster is)
    y_max : MU_MAX-MAG_AUTO threshold for linear regression
    mag_fit : value of MAG_AUTO to calculate the star line position after linear regression
    y_offsets : negative and positive offsets from star_line for the selection box
    mag_bounds : MAG_AUTO bounds for the selection box
    snr_min : minimum value for SNR_WIN to select the sources
    plot_mag_bounds : bounds for MAG_AUTO for the plots
    plot_mu_bounds : bounds for MU_MAX for the plots
    plot_y_bounds : bounds for (MU_MAX-MAG_AUTO) for the plots
    plot_rad_bound : maximum value for FLUX_RADIUS for the plots
    save_chckimg : parameter to save plots to check if everything worked well
    plot : parameter to display plots to check if everything worked well
    clean : delete the full catalog created by SExtractor at the end
    run_sex : run SExtractor. Can be set to False if you already have a FITS_LDAC catalog
    verbose : verbose parameter
    """
    study_name = '.'.join(detect_img.split('/')[-1].split('.')[:-1])
    if run_sex: run_sextractor(detect_img, weight_img, output_cat, config_folder, dir_chckimg, detect_thresh, analysis_thresh, verbose)
    hdul = fits.open(output_cat)
    data = hdul[2].data
    star_line = find_star_line(data, eps_DBSCAN, y_max, mag_fit, verbose, plot, plot_mag_bounds, plot_y_bounds, save_chckimg, f"{dir_chckimg}/starDetect_{study_name}.png")
    star_MUvMAG = MUvMAG_star_selection(data, star_line, y_offsets, mag_bounds, snr_min, plot, plot_mag_bounds, plot_y_bounds, save_chckimg, f"{dir_chckimg}/starLine_{study_name}.png")
    if plot | save_chckimg:
        star_selections = {'MUvMAG' : {'label': 'Stars (MU v. MAG)', 'color': 'r', 'flag': star_MUvMAG}}
        fig, ax = plt.subplots(1,2,figsize=(12,6))
        plot_MuvMAG(data, star_selections, plot_mag_bounds, plot_mu_bounds, ax_custom=ax[0])
        plot_SNR_radius(data, star_selections, plot_rad_bound, ax_custom=ax[1])
        if save_chckimg : fig.savefig(f"{dir_chckimg}/star_{study_name}.png", bbox_inches='tight', dpi=100)
        if plot : plt.show()
    if clean: os.remove(output_cat)
    if clean: os.remove(f"{dir_chckimg}/{study_name}_seg.fits")
    save_catalog(hdul, star_MUvMAG, output_cat_star)
    hdul.close()

def extract_stars_catalog(detect_img        : str, 
                          weight_img        : str,
                          output_cat        : str, 
                          output_cat_star   : str,
                          input_cat_star    : str, 
                          config_folder     : str,
                          dir_chckimg       : str,
                          detect_thresh     : float = 2.5,
                          analysis_thresh   : float = 2.5,
                          max_sep           : u.Quantity = 1.0*u.arcsec,
                          plot_mag_bounds   : tuple = (18,30),
                          plot_mu_bounds    : tuple = (13,25),
                          plot_rad_bound   : tuple = 20, 
                          save_chckimg      : bool = True,
                          plot              : bool = False,
                          clean             : bool = True,
                          run_sex           : bool = True,
                          verbose           : bool = False) -> None:
    """
    Create a point-like sources catalog from an image and a previous
    point-like sources catalog (e.g. from another band)to be used 
    in PSFEx to create the PSF of a telescope.
    
    detect_img : filename of the detection image to use
    weight_img : filename of the weight map
    output_cat : filename of the output full catalog
    output_cat_star : filename of the output point-like sources catalog
    input_cat_star : filename of the input point-like sources catalog
    config_folder : path to the SEtractor config filed folder
    dir_chckimg : path to folder to save checkimages (will be created if non-existing)
    detect_thresh : value of the DETECT_THRESH to use
    analysis_thresh : value of the ANALYSIS_THRESH to use
    max_sep : maximum angular separation for sources cross-match
    plot_mag_bounds : bounds for MAG_AUTO for the plots
    plot_mu_bounds : bounds for MU_MAX for the plots
    plot_rad_bound : maximum value for FLUX_RADIUS for the plots
    save_chckimg : parameter to save plots to check if everything worked well
    plot : parameter to display plots to check if everything worked well
    clean : delete the full catalog created by SExtractor at the end
    run_sex : run SExtractor. Can be set to False if you already have a FITS_LDAC catalog
    verbose : verbose parameter
    """
    study_name = '.'.join(detect_img.split('/')[-1].split('.')[:-1])
    if run_sex: run_sextractor(detect_img, weight_img, output_cat, config_folder, dir_chckimg, detect_thresh, analysis_thresh, verbose)
    if verbose: print ("Opening star catalogs")
    hdul = fits.open(output_cat, memmap=True, mode='denywrite') 
    data = hdul[2].data
    with fits.open(input_cat_star, memmap=True, mode='denywrite') as hdul_star: 
        data_star = hdul_star[2].data
    coord = SkyCoord(data['ALPHA_J2000'][:], data['DELTA_J2000'][:], unit='deg', frame='icrs')
    coord_star = SkyCoord(data_star['ALPHA_J2000'], data_star['DELTA_J2000'], unit='deg', frame='icrs')
    if verbose: print("Matching stars")
    idx, d2d, _ = coord_star.match_to_catalog_sky(coord)
    match = idx[d2d<max_sep]
    if plot | save_chckimg:
        print("Plotting")
        star_selections = {'MUvMAG' : {'label': 'Stars (MU v. MAG)', 'color': 'r', 'flag': match}}
        fig, ax = plt.subplots(1,2,figsize=(12,6))
        plot_MuvMAG(data, star_selections, plot_mag_bounds, plot_mu_bounds, ax_custom=ax[0])
        plot_SNR_radius(data, star_selections, plot_rad_bound, ax_custom=ax[1])
        if save_chckimg : fig.savefig(f"{dir_chckimg}/star_{study_name}.png", bbox_inches='tight', dpi=100)
        if plot : plt.show()
    if clean: os.remove(output_cat)
    save_catalog(hdul, match, output_cat_star)
    hdul.close()
    # hdul_star.close()


def main():
    pass
    
if __name__=="__main__":
    main()


import importlib.resources
import os
import re
import fnmatch
import glob
from typing import Union
import numpy as np
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.font_manager import FontProperties
from astropy.io import fits, ascii
from astropy.nddata import Cutout2D
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from astropy.coordinates import SkyCoord
from astropy.visualization import ImageNormalize, MinMaxInterval, ZScaleInterval, LogStretch
from astropy.visualization.wcsaxes import add_scalebar
import astropy.units as u

# Color dictionaries to make the plot nicer with colored filter names
color_dict = {'F090W':'#0000E3', 
              'F115W':'#1020FF',
              'F150W':'#0162FF',
              'F182M':'#00B1FF',
              'F200W':'#04D1FF',
              'F210M':'#02E8F4',
              'F250M':'#55FFAE',
              'F277W':'#87FF7F',
              'F300M':'#AAFF57',
              'F335M':'#EBFF0C',
              'F356W':'#FFD807',
              'F410M':'#FF6D03',
              'F430M':'#FF3F00',
              'F444W':'#FF330C',
              'F460M':'#DF0606',
              'F480M':'#B20000',}
channel_dict = {'F090W':'SW', 'F115W':'SW', 'F150W':'SW', 'F182M':'SW', 'F200W':'SW', 'F210M':'SW',
                'F250M':'LW', 'F277W':'LW', 'F300M':'LW', 'F335M':'LW', 'F356W':'LW', 'F410M':'LW', 'F430M':'LW', 'F444W':'LW', 'F460M':'LW', 'F480M':'LW'}
channel_color_dict = {'SW' : '#0162FF', 'LW' : '#FF330C'}

filters_waveband = {'F090W': {'pivot': 0.901, 'band': 0.194},
                    'F115W': {'pivot': 1.154, 'band': 0.225},
                    'F150W': {'pivot': 1.501, 'band': 0.318},
                    'F182M': {'pivot': 1.845, 'band': 0.238},
                    'F200W': {'pivot': 1.990, 'band': 0.461},
                    'F210M': {'pivot': 2.093, 'band': 0.205},
                    'F250M': {'pivot': 2.503, 'band': 0.181},
                    'F277W': {'pivot': 2.786, 'band': 0.672},
                    'F300M': {'pivot': 2.996, 'band': 0.318},
                    'F335M': {'pivot': 3.365, 'band': 0.347},
                    'F356W': {'pivot': 3.563, 'band': 0.787},
                    'F410M': {'pivot': 4.092, 'band': 0.436},
                    'F430M': {'pivot': 4.280, 'band': 0.228},
                    'F444W': {'pivot': 4.421, 'band': 1.024},
                    'F460M': {'pivot': 4.624, 'band': 0.228},
                    'F480M': {'pivot': 4.834, 'band': 0.303},}
# Source : https://jwst-docs.stsci.edu/jwst-near-infrared-camera/nircam-instrumentation/nircam-filters

def add_good_scalebar(ax, wcs, color='white', fraction=0.25, corner='bottom right', pad=0.1, fontsize='x-large'):
    """
    Utility function for astropy.visualization.wcsaxes.add_scalebar
    to automaticaly resize the scalebar to the size of the image.

    ax : matplotlib.pyplot.Axes where to draw the scale bar
    wcs : WCS of the image
    color : color of the scale bar
    fraction : fraction of the image to aim for the scale bar (will always be smaller)
    corner : where to place the scale bar
    pad : padding around the scale bar
    fontsize : fontsize of the scale text
    """
    width = abs(proj_plane_pixel_scales(wcs)[0]*wcs.pixel_shape[0])*u.degree
    good_values = [0.5*u.arcsec, 1*u.arcsec, 2*u.arcsec, 3*u.arcsec, 5*u.arcsec, 10*u.arcsec, 20*u.arcsec,
                0.5*u.arcmin, 1*u.arcmin, 2*u.arcmin, 3*u.arcmin, 5*u.arcmin, 10*u.arcmin, 20*u.arcmin,
                0.5*u.degree, 1*u.degree, 2*u.degree, 5*u.degree, 10*u.degree]
    dist = np.array([(fraction*width-val).value for val in good_values])
    dist[dist<0] = np.inf
    size = good_values[np.argmin(dist)]
    add_scalebar(ax, size, label=f"{size:latex}", color=color, fontproperties=FontProperties(size=fontsize), label_top=True, pad=pad, corner=corner)

def save_cutouts(generic_filename   : str, 
                 center             : Union[tuple[str],SkyCoord] , 
                 size               : Union[int,u.Quantity],
                 save_folder        : str = None, 
                 plot               : bool = False,
                 plot_str           : str = None,
                 verbose            : bool = False) -> None :
    """
    Create cutouts for all selected images, 
    centered on the same point and with the same size.

    * `generic_filename` (`str` with wildcard) : 
        Generic filename for the images to process.
        Can use wildcards (*,?) readable by `glob.glob`.
    * `center` (`tuple[str]` or `SkyCoord`) :
        Center of the cutouts.
        Can be in pixel units, or an `astropy.coordinates.SkyCoord` object (e.g. RA-DEC).
    * `size` (`int` or `Quantity`):
        Size of the cutouts.
        Can be in pixel units, or an `astropy.units.Quantity` object (e.g. angular size).
        WARNING : If pixel scale is different between images, giving a size in pixel units will result in non-aligned images.
    * `save_folder` (`str`, optional):
        Folder to save the cutouts in.
        By default, the images will be saved in a `cutout` sub_folder of the initial folder.
    * `plot` (`bool`, optional):
        Plots the cutouts with MinMax and Log stretching.
    * `plot_str` (`str`, optional):
        String to be found in image name to plot it. 
        Only active if `plot=True`. By default, all images are plotted.
    * `verbose` (`bool`, optional)
    """
    image_list = glob.glob(generic_filename)
    n = len(image_list)
    if verbose: print(f"Number of images found : {n}")
    if verbose: print(f"Images found :")
    if verbose: 
        for img in image_list: print(img)
    if plot: 
        len_plot = np.sum([fnmatch.fnmatch(img, plot_str) for img in image_list])
        size_plot = 1
        while size_plot*size_plot < len_plot: size_plot += 1
        fig, j = plt.figure(figsize=(3*size_plot,3*size_plot)), 1
    for i,img in enumerate(image_list):
        print(f"Image {i+1} / {n}")
        # Load image
        hdu = fits.open(img, memmap=True)[0]
        wcs = WCS(hdu.header)
        # Cutout
        cutout = Cutout2D(hdu.data, 
                    position=center, size=size, 
                    wcs=wcs,
                    mode='partial', fill_value=0.0)
        # Save cutout
        hdu.data = cutout.data
        hdu.header.update(cutout.wcs.to_header())
        parsed_name = img.split("/")
        folder = "/".join(parsed_name[:-1])
        init_name = parsed_name[-1]
        name = f"{'.'.join(init_name.split('.')[:-1])}_cutout.fits"
        if save_folder is None: save_folder = f"{folder}/cutout"
        if not os.path.exists(save_folder): os.mkdir(save_folder) 
        hdu.writeto(f"{save_folder}/{name}", overwrite=True)
        # Plot cutout
        if plot:
            if plot_str is None or fnmatch.fnmatch(img, plot_str):
                ax = fig.add_subplot(size_plot,size_plot,j, projection=cutout.wcs)
                norm = ImageNormalize(cutout.data, interval=ZScaleInterval())
                ax.imshow(cutout.data, cmap='gray', origin='lower', norm=norm)
                ax.set_title(init_name, size='xx-small')
                ax.set_axis_off()
                add_good_scalebar(ax, cutout.wcs)
                j+=1
    if plot: fig.tight_layout()
    if plot: plt.show()
    if plot: return fig

def show_source(id, cat, filter_list,
                data_folder,
                model_folder,
                resid_folder,
                data_suffix="sci.fits",
                model_suffix="sci_1.fits",
                resid_suffix="sci_1.fits",
                fov=3*u.arcsec, 
                ra_name='world_centroid_alpha', de_name='world_centroid_delta',):
    """
    Displays data, model and residual images for each band,
    for the source at a given index in a given catalog.

    id : index of the source in cat
    cat : catalog from SE++
    filter_list : list of the filters (in lowercase) to use for the display
    fov : field of view of the cutouts around the source.
          Must be a astropy.units.Quantity
    ra/de_name : column name in cat for right ascension and declination
    data/model/resid_folder : folder to the data/model/resid images
    data/model/resid_suffix : suffix for image name selection (can contain glob.glob wildcards)
    """
    target = SkyCoord(cat[ra_name][id], cat[de_name][id], frame='icrs', unit='deg')
    size = (1.1*fov,1.1*fov)

    fig = plt.figure(figsize=(2*len(filter_list)+1,2*3+0.5))
    for i, filter in enumerate(filter_list):
        # Loads images
        hdu       = fits.open(glob.glob(f"{data_folder}/*{filter}*{data_suffix}")[0], memmap=True)[0]
        model_hdu = fits.open(glob.glob(f"{model_folder}/model*{filter}*{model_suffix}")[0], memmap=True)[0]
        resid_hdu = fits.open(glob.glob(f"{resid_folder}/resid*{filter}*{resid_suffix}")[0], memmap=True)[0]
        wcs = WCS(hdu.header)
        # Cutouts in images
        cutout       = Cutout2D(hdu.data,       position=target, size=size, wcs=wcs, mode='partial', fill_value=0.0)
        model_cutout = Cutout2D(model_hdu.data, position=target, size=size, wcs=wcs, mode='partial', fill_value=0.0)
        resid_cutout = Cutout2D(resid_hdu.data, position=target, size=size, wcs=wcs, mode='partial', fill_value=0.0)
        # Plots data
        ax = fig.add_subplot(3,len(filter_list),i+1, projection=cutout.wcs)
        norm = ImageNormalize(cutout.data, interval=ZScaleInterval())
        # norm = ImageNormalize(cutout.data, interval=MinMaxInterval(), stretch=LogStretch())
        ax.imshow(cutout.data, cmap='gray', origin='lower', norm=norm)
        add_good_scalebar(ax, cutout.wcs, fraction=1.0, corner='bottom')
        ax.set_title(filter.upper(), fontsize=24)
        ax.set_axis_off()
        # Plots model
        ax_model = fig.add_subplot(3,len(filter_list),i+len(filter_list)+1, projection=cutout.wcs)
        ax_model.imshow(model_cutout.data, cmap='gray', origin='lower', norm=norm)
        ax_model.set_axis_off()
        # Plots residual
        ax_resid = fig.add_subplot(3,len(filter_list),i+2*len(filter_list)+1, projection=cutout.wcs)
        ax_resid.imshow(resid_cutout.data, cmap='gray', origin='lower', norm=norm)
        ax_resid.set_axis_off()
    plt.figtext(1.0, 0.79, "Data", figure=fig, va='center', ha='left', rotation='vertical', fontsize='xx-large')
    plt.figtext(1.0, 0.48, "Model", figure=fig, va='center', ha='left',rotation='vertical', fontsize='xx-large')
    plt.figtext(1.0, 0.17, "Residual", figure=fig, va='center', ha='left',rotation='vertical', fontsize='xx-large')
    plt.figtext(0.0, 0.48, f"Source #{id}", figure=fig, va='center', ha='right', rotation='vertical', fontsize=24)
    fig.tight_layout()
    plt.show()
    return fig

def plot_filters(ax, filter_list=None, scale=1, names=True, 
                 throughput_folder="data/NIRCam_Filters-Throughput"):
    """
    Plots normalized filter throughputs on ax
    Source : https://jwst-docs.stsci.edu/jwst-near-infrared-camera/nircam-instrumentation/nircam-filters

    ax : matplotlib.pyplot.Axes to plot on
    filter_list : list of the filters to plot
    scale : maximum value of the plotted throughput curves
    names : whether to display the names of the filters
    throughput_folder : folder with the throughput data downloaded from the JWST documentation
    """
    throughput_folder = importlib.resources.files('dja_sepp').joinpath(throughput_folder)
    filter_list = [filter.lower() for filter in color_dict.keys()] if filter_list is None else filter_list
    for filter in filter_list:
        table = ascii.read(f"{throughput_folder}/{filter.upper()}_mean_system_throughput.txt")
        color = color_dict[filter.upper()] if filter.upper() in color_dict else '#000000'
        max_val = scale if 'w' in filter else scale/2
        ax.plot(table['Microns'], max_val*table['Throughput']/table['Throughput'].max(), color=color)
        ax.fill_between(table['Microns'], max_val*table['Throughput']/table['Throughput'].max(), alpha=0.4, color=color)
        if names: ax.text(np.average(table['Microns'], weights=table['Throughput']), 0.55*max_val, filter.upper(), ha='center', va='center', c='k', size='x-small')

def plot_photometric_spectrum(id, cat, filter_list, mag=False, custom_ax=None, title=None):
    """
    Plots photometric spectrum of cat[id] in flux or mag depending of `mag` (default, flux).
    The value of the wavelength and bandwidth associated to each band comes from the JWST documentation.
    Source : https://jwst-docs.stsci.edu/jwst-near-infrared-camera/nircam-instrumentation/nircam-filters

    id : index of the source in cat
    cat : catalog from SE++
    filter_list : list of the filters to plot
    mag : whether to plot the spectrum in AB mag or flux (default)
    custom_ax : matplotlib.pyplot.Axes to plot on.
                By default, the function creates a new figure
    title : title to set for the plot
    """

    fig, ax = plt.subplots(figsize=(12,3)) if custom_ax is None else (None, custom_ax)
    # DJA images pixel value are in 10 nJy
    flux_target = [10*cat[id-1][f"FLUX_MODEL_{filter.upper()}"] for filter in filter_list]
    flux_err_target = [10*cat[id-1][f"FLUX_MODEL_{filter.upper()}_err"] for filter in filter_list]
    mag_target = [cat[id-1][f"MAG_MODEL_{filter.upper()}"] for filter in filter_list]
    mag_err_target = [cat[id-1][f"MAG_MODEL_{filter.upper()}_err"] for filter in filter_list]

    plot_filters(ax, filter_list, (max(mag_target) if mag else max(flux_target))/5)
    if mag:
        ax.errorbar([filters_waveband[filter.upper()]['pivot'] for filter in filter_list], mag_target, 
                    xerr=np.array([filters_waveband[filter.upper()]['band'] for filter in filter_list])/2, yerr=mag_err_target, 
                    fmt='o', ms=3, c='k', capsize=3, capthick=1)
    else:
        ax.errorbar([filters_waveband[filter.upper()]['pivot'] for filter in filter_list], flux_target, 
                    xerr=np.array([filters_waveband[filter.upper()]['band'] for filter in filter_list])/2, yerr=flux_err_target, 
                    fmt='o', ms=3, c='k', capsize=3, capthick=1)
    ax.set_xlabel("Wavelength (µm)")
    if mag:
        ax.set_ylabel("AB Mag")
    else:
        ax.set_ylabel(r"$F_{\nu}$ (nJy)")
    ax.set_title(title)
    if custom_ax is None: plt.show()
    if custom_ax is None: return fig

def plot_group_filter(filter_list, plot_func, **kwargs):
    """
    Make plots for different filters

    filter_list : list of filter names (in lowercase)
    plot_func : function defining what to plot for one filter.
            Its first two arguments must be a plt.Axes and the filter name
    **kwargs : additionnal arguments to pass to plot_func

    Returns : fig, axs of the figure
    """
    channel, count = np.unique([channel_dict[filter.upper()] for filter in filter_list], return_counts=True)
    channel_count = dict(zip(channel, count))
    w = min(5,channel_count[max(channel_count, key=channel_count.get)])
    h = sum([-(channel_count[channel]//-w) for channel in channel_count])
    fig, axs = plt.subplots(h,w,figsize=(3*w+1,3*h+1), sharex=True, sharey=True, gridspec_kw = {'wspace':0, 'hspace':0})
    indices = [i for i in range(len(axs.flatten()))]
    for i, filter in enumerate(filter_list):
        if (len(channel_count)==2) & (channel_dict[filter.upper()]=="LW"):
            mod = channel_count["SW"]%w
            mod = w if mod==0 else mod
            i += max(0, w-mod)
        indices.remove(i)
        ax = axs.flatten()[i] if type(axs)==np.ndarray else axs
        plot_func(ax, filter, **kwargs)
    for i in indices:
        axs.flatten()[i].set_axis_off()
    return fig, axs

def get_filter_list(keys):
    """
    Find filter list from column names of a SE++ catalog

    keys : list of column names (obtained throught cat.keys())

    Returns : list of filter names (in lowercase)
    """
    filter_list = [key.split("_")[-1].lower() for key in keys if (('MAG_MODEL' in key) and ('err' not in key) and ('BULGE' not in key) and ('DISK' not in key))]
    filter_list.sort()
    return filter_list

def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [ atoi(c) for c in re.split(r'(\d+)', text) ]


def main():
    fig, ax = plt.subplots(figsize=(12,5))
    plot_filters(ax, ['f090w', 'f115w', 'f200w', 'f277w', 'f356w', 'f444w'])
    plt.show()

if __name__=="__main__":
    main()


import os
import subprocess
import glob
import re
import sys
import boto3
from astropy.io import fits
import numpy as np
from . import utils

def find_images(generic_img : str,
                generic_wht : str,
                generic_psf : str,
                filter_list = None,
                verbose : bool = False):
    """Find the images, weight maps and PSFs according to generic names
    
    generic_img : generic filename for the images (can use *,? wildcards understandable by glob.glob)
    generic_wht : generic filename for the weight maps (can use *,? wildcards understandable by glob.glob)
    generic_psf : generic filename for the PSFs (can use *,? wildcards understandable by glob.glob)
    filter_list (optional) : list of filters to find in the file names.
        If set to None, will automatically be created by looking for f\d+\w+ Regex in images names
    verbose (default, False)
        
    Returns : list of images names, list of weight maps names, list of PSFs names"""
    images = glob.glob(generic_img)
    weights = glob.glob(generic_wht)
    psfs = glob.glob(generic_psf)
    if filter_list is None:
        filter_list = list(set([re.search('(f\d+\w+)', img).group(1) for img in images]))
        filter_list.sort()
    images = [[file for file in images if filter in file][0] for filter in filter_list]
    for i, img in enumerate(images):
        with fits.open(img, memmap=True) as hdul:
            if np.max(hdul[0].data)==0.0:
                blank_img = images.pop(i)
                blank_filter = filter_list.pop(i)
                print(f"Filter {blank_filter.upper()} is blank !! ({blank_img})")
    weights = [[file for file in weights if filter in file][0] for filter in filter_list]
    psfs = [[file for file in psfs if filter in file][0] for filter in filter_list]
    return images, weights, psfs

def run_sepp(detect_img : str,
             generic_img : str,
             generic_wht : str,
             generic_psf : str,
             output_catname : str,
             checkimg_path : str,
             python_config : str,
             filt : str,
             filter_list = None,
             fit_case : str = 'sersic_rg4',
             properties : str = 'PixelCentroid,WorldCentroid,SourceIDs,GroupInfo,GroupStamp,SourceFlags,NDetectedPixels,NCorePixel,AperturePhotometry,AutoPhotometry,FluxRadius,SNRRatio,ShapeParameters,FlexibleModelFitting',
             thread_count : int = 128,
             tile_memory_limit : int = 16384,
             tile_size : int = 1000,
             verbose : bool = True):
    """
    Run SourceXtractor++ on a given set of images
    
    detect_img : filename of the detection image to use
    generic_img : generic filename for the images (can use *,? wildcards understandable by glob.glob)
    generic_wht : generic filename for the weight maps (can use *,? wildcards understandable by glob.glob)
    generic_psf : generic filename for the PSFs (can use *,? wildcards understandable by glob.glob)
    output_catname : filename of the output catalog
    checkimg_path : path to the folder for the model and residual images
    python_config : filename of the Python configuration file
    filt : convolution filter filename to use for the detection
    filter_list (optional) : list of filters to find in the file names.
        If set to None, will automatically be created by looking for f\d+\w+ Regex in images names
    fit_case (default, sersic_rg4) : the model for fitting in SE++
    properties (default, complete) : SE++ properties to compute and export
    thread_count (default, 128) : number of threads to use for SE++
    tile_memory_limit (default, 16384) : maximum RAM allocated to SE++ (in MiB)
    tile_size (default, 1000) : size in px of the tiles created by SE++
    verbose (default, True)
    """
    images, weights, psfs = find_images(generic_img, generic_wht, generic_psf, filter_list, verbose)
    args = {'fit_case' : fit_case,
            'list_of_IMG_names' : images,
            'list_of_WHT_names' : weights,
            'list_of_PSF_names' : psfs
            }
    os.makedirs(checkimg_path, exist_ok=True)
    os.environ['MKL_NUM_THREADS'] = '1'
    os.environ['MKL_DYNAMIC'] = 'FALSE'
    os.environ['OMP_NUM_THREADS'] = '1'
    os.environ['OMP_DYNAMIC'] = 'FALSE'
    subprocess.run(f"sourcextractor++ \
                   --python-config-file {python_config} \
                   --python-arg '{args}' \
                   --output-catalog-filename {output_catname} \
                   --output-properties {properties} \
                   --output-flush-size 10 \
                   --detection-image {detect_img} \
                   --weight-type WEIGHT \
                   --weight-absolute 0 \
                   --segmentation-filter {filt} \
                   --background-cell-size 128 \
                   --segmentation-algorithm LUTZ \
                   --smoothing-box-size 5 \
                   --detection-threshold 0.80 \
                   --detection-minimum-area 7 \
                   --partition-corethreshold yes \
                   --core-threshold-value 1.5 \
                   --core-minimum-area 9 \
                   --grouping-algorithm split \
                   --partition-multithreshold yes \
                   --partition-minimum-area 18 \
                   --grouping-hard-limit 30 \
                   --partition-minimum-contrast 0.0001 \
                   --partition-threshold-count 42 \
                   --use-cleaning yes \
                   --cleaning-minimum-area 15 \
                   --model-fitting-iterations 550 \
                   --sampling-scale-factor 1 \
                   --sampling-max-fit-size 500 \
                   --check-image-model-fitting {checkimg_path}/model.fits \
                   --check-image-residual {checkimg_path}/resid.fits \
                   --log-file {output_catname.replace('.fits', '.log')} \
                   --log-level DEBUG \
                   --thread-count {thread_count} \
                   --tile-memory-limit {tile_memory_limit} \
                   --tile-size {tile_size} \
                   ",
                   shell=True)

def main_tile(tile, img_dir, sepp_dir, psf_dir, config_dir):
    # Python configuration file for SE++
    python_config=f'{config_dir}/sepp-config.py'
    # Detection image
    detect_img=glob.glob(f'{img_dir}/*ir*sci*-{tile}.fits')[0]
    # Filter to use on the detection image
    filt=f'{config_dir}/gauss_1.5_3x3.conv'
    # Output
    name = ".".join(detect_img.split("/")[-1].split(".")[:-1])
    output_catname=f'{sepp_dir}/{name}_sepp_cat.fits'
    checkimg_path=f'{sepp_dir}/checkimages'
    os.makedirs(checkimg_path, exist_ok=True)
    # Run SE++
    run_sepp(detect_img=detect_img,
             generic_img=f"{img_dir}/*clear*sci*-{tile}.fits",
             generic_wht=f"{img_dir}/*clear*wht*-{tile}.fits",
             generic_psf=f"{psf_dir}/*star_psf.psf",
             output_catname=output_catname,
             checkimg_path=checkimg_path,
             python_config=python_config,
             filt=filt,
             filter_list = None,
             fit_case='sersic_rg4',
             properties='PixelCentroid,WorldCentroid,SourceIDs,GroupInfo,GroupStamp,SourceFlags,NDetectedPixels,NCorePixel,AperturePhotometry,AutoPhotometry,FluxRadius,SNRRatio,ShapeParameters,FlexibleModelFitting',
             thread_count=128,
             tile_memory_limit=16384,
             tile_size=2000,
             verbose=True)
    # Save to S3
    utils.save_s3(output_catname, 'aurelien_sepp', 'sepp/GDS/tiles')

def main_run(img_dir, sepp_dir, psf_dir, config_dir):
    # Python configuration file for SE++
    python_config=f'{config_dir}/sepp-config.py'
    # Detection image
    detect_img=glob.glob(f'{img_dir}/*ir*sci*.fits')[0]
    # Filter to use on the detection image
    filt=f'{config_dir}/gauss_1.5_3x3.conv'
    # Output
    name = ".".join(detect_img.split("/")[-1].split(".")[:-1])
    output_catname=f'{sepp_dir}/{name}_sepp_cat.fits'
    checkimg_path=f'{sepp_dir}/checkimages'
    os.makedirs(checkimg_path, exist_ok=True)
    # Run SE++
    run_sepp(detect_img=detect_img,
             generic_img=f"{img_dir}/*clear*sci*.fits",
             generic_wht=f"{img_dir}/*clear*wht*.fits",
             generic_psf=f"{psf_dir}/*star_psf.psf",
             output_catname=output_catname,
             checkimg_path=checkimg_path,
             python_config=python_config,
             filt=filt,
             filter_list = None,
             fit_case='sersic_rg4',
             properties='PixelCentroid,WorldCentroid,SourceIDs,GroupInfo,GroupStamp,SourceFlags,NDetectedPixels,NCorePixel,AperturePhotometry,AutoPhotometry,FluxRadius,SNRRatio,ShapeParameters,FlexibleModelFitting',
             thread_count=128,
             tile_memory_limit=16384,
             tile_size=2000,
             verbose=True)

def main():
    tile = sys.argv[1]
    img_dir = sys.argv[2] if len(sys.argv)>2 else "/home/ec2-user/DAWN/DJA-SEpp/image/GDS/cutout/tiles"
    sepp_dir = sys.argv[3] if len(sys.argv)>3 else "/home/ec2-user/DAWN/DJA-SEpp/sepp/GDS/tiles"
    psf_dir = sys.argv[4] if len(sys.argv)>4 else "/home/ec2-user/DAWN/DJA-SEpp/psfex/GDS"
    config_dir = sys.argv[5] if len(sys.argv)>5 else "/home/ec2-user/DAWN/DJA-SEpp/config"
    main_tile(tile, img_dir, sepp_dir, psf_dir, config_dir)

if __name__=='__main__':
    main()


from .utils import color_dict, channel_dict, channel_color_dict, filters_waveband, add_good_scalebar, save_cutouts, show_source, plot_filters, plot_photometric_spectrum, plot_group_filter, get_filter_list
from .s3 import find_files, decompress_save, decompress_save_to_S3, save_s3
from .sextractor import run_sextractor, show_vignets, plot_MuvMAG, hist_CLASS_STAR, plot_SNR_radius, plot_MuvMAG_manual, find_star_line, MUvMAG_star_selection, save_catalog, extract_stars, extract_stars_catalog
from .psfex import run_psfex, compare_star
from .sepp import find_images, run_sepp, main_tile, main_run
from .tiles import create_tiles, tile_positions, tile_grid, save_tiles, plot_tiles, batch_tiling, merge_catalogs, merge_tiles, merge_images


import os
import re
import fnmatch
import glob
import numpy as np
import numpy.ma as ma
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from astropy.io import fits
from astropy.nddata import Cutout2D
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales, pixel_to_skycoord
from astropy.visualization import ImageNormalize, MinMaxInterval, ZScaleInterval, LogStretch
from astropy.coordinates import SkyCoord
from astropy.table import vstack
import astropy.units as u
from reproject import reproject_interp, reproject_exact
from reproject.mosaicking import find_optimal_celestial_wcs, reproject_and_coadd
from . import utils

def create_tiles(data, wcs, centers, sizes):
    """
    Create the tiles out of a big image and given center position and sizes of tiles

    data : data of the main image
    wcs : WCS of the main image
    centers : list of SkyCoord for the center of each tile
    sizes : list of tuple of angular sizes for each tile

    Returns : list of Cutout2D representing the tiles
    """
    tiles = []
    for i in range(len(centers)):
        cutout = Cutout2D(data, position=centers[i], size=sizes[i], wcs=wcs, mode='partial', fill_value=0.0)
        tiles.append(cutout)
    return tiles

def tile_positions(wcs, nx, ny, overlap):
    """
    Calculates the centers and sizes of tiles based on a grid

    wcs : WCS of the main image
    nx/y : number of tiles on the x/y axis
    overlap : angular size of the overlap between tiles

    Returns : lists of centers and sizes for the tiles
    """
    width, height = wcs.pixel_shape
    pixscale = abs(proj_plane_pixel_scales(wcs)[0])*u.degree
    overlap_px = int(overlap/pixscale)
    width_tile = (width-overlap_px)/nx + overlap_px
    height_tile = (height-overlap_px)/ny + overlap_px
    centers = []
    sizes = []
    for j in range(ny):
        for i in range(nx):
            center = pixel_to_skycoord(0.5*width_tile + i*(width_tile-overlap_px), 0.5*height_tile + j*(height_tile-overlap_px), wcs)
            centers.append(center)
            sizes.append((width_tile*pixscale, height_tile*pixscale))
    return centers, sizes

def tile_grid(wcs, max_size_x, max_size_y, overlap):
    """
    Calculates the grid size for a given maximum angular tile size

    wcs : WCS of the main image
    max_size_x/y : maximum angular size on the x/y axis for the tiles (without overlap)
    overlap : angular size of the overlap between tiles

    Returns : number of tiles on the x and y axis
    """
    width_px, height_px = wcs.pixel_shape
    pixscale = abs(proj_plane_pixel_scales(wcs)[0])*u.degree
    width, height = width_px*pixscale, height_px*pixscale
    nx, ny = int(width/max_size_x)+1, int(height/max_size_y)+1
    return nx, ny

def save_tiles(tiles, main_filename, save_folder=None, verbose=False):
    """
    Saves the tiles as FITS images

    tiles : list of Cutout2D representing the tiles
    main_filename : filename of the main image
    save_folder : folder to save the cutouts in.
        By default, tiles are saved in a `cutout` sub-folder where the main image is.
    """
    hdu = fits.open(main_filename, memmap=True)[0]
    for i, tile in enumerate(tiles):
        if verbose: print(f"Tile {i+1} / {len(tiles)}")
        hdu.data = tile.data
        hdu.header.update(tile.wcs.to_header())
        parsed_name = main_filename.split("/")
        folder = "/".join(parsed_name[:-1])
        init_name = parsed_name[-1]
        name = f"{'.'.join(init_name.split('.')[:-1])}_tile-{i}.fits"
        if save_folder is None: save_folder = f"{folder}/tiles"
        if not os.path.exists(save_folder): os.mkdir(save_folder) 
        hdu.writeto(f"{save_folder}/{name}", overwrite=True)

def plot_tiles(nx, ny, tiles, plot_main=False, data=None, wcs=None):
    """
    Plot the tiles for visual inspection

    nx/y : number of tiles on the x/y axis
    tiles : list of Cutout2D representing the tiles
    plot_main : whether to plot the main image or not
    data : data array of the main image (required if plot_main==True)
    wcs : WCS of the main image (required if plot_main==True)
    """
    h, w = tiles[0].data.shape
    plot_x_size = min(12, 2*nx)
    plot_y_size = min(8, plot_x_size*ny/nx*h/w)
    plot_x_size = min(12, plot_y_size*nx/ny*w/h)
    if plot_main:
        fig = plt.figure(layout='constrained', figsize=(plot_x_size,plot_y_size/2))
        gs = GridSpec(ny, nx*2, figure=fig)
        ax_main = fig.add_subplot(gs[:,nx:], projection=wcs)
        norm = ImageNormalize(data, interval=ZScaleInterval())
        ax_main.imshow(data, cmap='gray', origin='lower', norm=norm)
        utils.add_good_scalebar(ax_main, wcs)
        ax_main.set_xlabel('Right Ascension (J2000)')
        ax_main.set_ylabel('Declination (J2000)')
    else:
        fig = plt.figure(layout='constrained', figsize=(plot_x_size,plot_y_size))
        gs = GridSpec(ny, nx, figure=fig)
        norm = ImageNormalize(tiles[0].data, interval=ZScaleInterval())
    for j in range(ny):
        for i in range(nx):
            tile = tiles[i+nx*j]
            ax = fig.add_subplot(gs[ny-j-1,i], projection=tile.wcs)
            ax.imshow(tile.data, cmap='gray', origin='lower', norm=norm)
            ax.set_axis_off()
            if (j==ny-1) & (i==nx-1):
                utils.add_good_scalebar(ax, tile.wcs)
    return fig

def batch_tiling(generic_filename   : str,
                 tile_max_size      : u.Quantity,
                 overlap            : u.Quantity,
                 save_folder        : str = None, 
                 plot               : bool = False,
                 plot_str           : str = None,
                 verbose            : bool = False) -> None :
    """
    Create tiles for all selected images, 
    centered on the same points and with the same angular size.

    * `generic_filename` (`str` with wildcard) : 
        Generic filename for the images to process.
        Can use wildcards (*,?) readable by `glob.glob`.
    * `tile_max_size` (`u.Quantity`) :
        Maximum angular size of the tiles (without overlap)
    * `overlap` (`u.Quantity`) :
        Angular size of overlap between tiles
    * `save_folder` (`str`, optional):
        Folder to save the cutouts in.
        By default, the images will be saved in a `cutout` sub-folder of the initial folder.
    * `plot` (`bool`, optional):
        Plots the tiles for the first band with MinMax and Log stretching.
    * `plot_str` (`str`, optional):
        Regex to choose the image to plot. Can use wildcards (*,?) readable by `glob.glob`.
    * `verbose` (`bool`, optional)
    """
    image_list = glob.glob(generic_filename)
    n = len(image_list)
    if verbose: print(f"Number of images found : {n}")
    if verbose: print(f"Images found :")
    if verbose: 
        for img in image_list: print(img)
    for i,img in enumerate(image_list):
        print(f"Image {i+1} / {n}")
        # Load image
        hdu = fits.open(img, memmap=True)[0]
        wcs = WCS(hdu.header)
        data = hdu.data
        # Calculate tiling on first image
        if i==0:
            nx, ny = tile_grid(wcs, tile_max_size, tile_max_size, overlap)
            centers, sizes = tile_positions(wcs, nx, ny, overlap)
        # Tiles the image
        tiles = create_tiles(data, wcs, centers, sizes)
        # Save tiles
        save_tiles(tiles, img, save_folder, verbose=verbose)
        if plot & fnmatch.fnmatch(img, plot_str):
            fig = plot_tiles(nx, ny, tiles, plot_main=False)
    if plot: plt.show()
    if plot: return fig

def merge_catalogs(cat1, cat2, max_sep=0.25*u.arcsec, filter_merge='f200w', return_matches=False):
    """
    Merge two tile catalog by finding matches and selecting between these matches by lowest mag error

    cat1/cat2 : catalogs (astropy.table.Table) to merge
    max_sep : maximum separation between sources for cross match
    filter_merge : filter to use for the selection using lowest mag error
    return_matches : whether to return the indices of the sources in cat1/cat2 matched

    Returns : merged catalog (astropy.table.Table)
    """
    # Creating columns that are in one catalog but not the other, and in list-columns (missing filters)
    keys1 = cat1.keys()
    filter_list1 = utils.get_filter_list(keys1)
    keys2 = cat2.keys()
    filter_list2 = utils.get_filter_list(keys2)
    filter_list = list(set(filter_list1 + filter_list2))
    filter_list.sort()
    list_cols = []
    for key in cat1.keys():
        if len(cat1[key].shape)>1:
            if cat1[key].shape[1]==len(filter_list1):
                list_cols.append(key)
    for i, filter in enumerate(filter_list):
        if filter not in filter_list1:
            for col in list_cols:
                cat1[col] = np.insert(ma.array(cat1[col]), i, ma.masked, axis=1)
        if filter not in filter_list2:
            for col in list_cols:
                cat2[col] = np.insert(ma.array(cat2[col]), i, ma.masked, axis=1)
    # Matching sources
    coord1 = SkyCoord(cat1['world_centroid_alpha']*u.degree, cat1['world_centroid_delta']*u.degree)
    coord2 = SkyCoord(cat2['world_centroid_alpha']*u.degree, cat2['world_centroid_delta']*u.degree)
    idx, d2d, _ = coord1.match_to_catalog_sky(coord2)
    match1 = np.where(d2d<max_sep)[0]
    match2 = idx[match1]
    # Selecting matched sources based on mag uncertainty
    merge_choice = cat1[match1][f'MAG_MODEL_{filter_merge.upper()}_err']<cat2[match2][f'MAG_MODEL_{filter_merge.upper()}_err']
    merged_1 = cat1[match1][merge_choice]
    merged_2 = cat2[match2][~merge_choice]
    # Selecting unmatched sources
    cat1_unmatched = cat1[[i for i in range(len(cat1)) if i not in match1]]
    cat2_unmatched = cat2[[i for i in range(len(cat2)) if i not in match2]]
    # Merging catalogs
    merged = vstack([cat1_unmatched, cat2_unmatched, merged_1, merged_2], join_type='outer')
    if return_matches: return merged, match1, match2
    return merged
    # return cat1

def merge_tiles(catalogs, max_sep=0.25*u.arcsec, filter_merge='f200w'):
    """
    Merge tile catalogs by finding matches and selecting between these matches by lowest mag error

    catalogs : catalogs (astropy.table.Table) to merge
    max_sep : maximum separation between sources for cross match
    filter_merge : filter to use for the selection using lowest mag error

    Returns : merged catalog (astropy.table.Table)
    """
    merged = catalogs[0]
    for i in range(1, len(catalogs)):
        merged = merge_catalogs(catalogs[i], merged, max_sep, filter_merge)
    return merged

def merge_images(folder, filter_list, type, suffix="*sci*tile-[!full]*", wcs=None, shape=None, exact=False, out_folder=None, verbose=False):
    """
    Mosaic tile images (data, model, residual) by reprojection and co-addition

    folder : folder of the images to mosaic
    filter_list : list of filters (in image names) to use for mosaicing process
    type : type of images to mosaic (fromerly, prefix of image name)
    suffix : suffix of image_name (after filter)
    wcs : WCS for final image
    shape : shape for final image
    exact : use `reproject.reproject_exact` for reprojection (by default, `reproject.reproject_interp`)
    out_folder : folder to save the mosaiced images. By default, same folder as `folder`
    verbose : verbose
    """
    if (wcs is None) or (shape is None):
        if verbose: print("Finding optimal WCS")
        images = []
        for filter in filter_list:
            images.extend(glob.glob(f"{folder}/*{type}*{filter}*{suffix}*"))
        wcs, shape = find_optimal_celestial_wcs(images, auto_rotate=True)
    for filter in filter_list:
        if verbose: print(f"---- {filter.upper()} ----")
        images = glob.glob(f"{folder}/*{type}*{filter}*{suffix}*")
        if verbose: print(f"{filter.upper()} : {images}")
        if verbose: print("Mosaicing")
        reproject_function = reproject_exact if exact else reproject_interp
        mosaic, footprint = reproject_and_coadd(images, wcs, shape_out=shape, 
                                                reproject_function=reproject_function,
                                                match_background=False)
        if verbose: print("Saving")
        full = fits.PrimaryHDU(mosaic, wcs.to_header())
        full.data = full.data.astype('float32')
        out_folder = folder if out_folder is None else out_folder
        name = re.sub('tile-\d+', 'tile-full', images[0].split('/')[-1])
        full.writeto(f"{out_folder}/{name}", overwrite=True)
    return wcs, shape


import os
import subprocess
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import SymLogNorm
from astropy.io import fits

# To work, this file requires you to have PSFEx installed on your machine.

def run_psfex(cat_name, 
              output_cat, 
              dir_chckimg, dir_psfex, 
              config_file, 
              fwhm_range='1.0,6.0', min_snr=200, 
              save_psf_png=False, verbose=False):
    """
    Run PSFEx on the FITS_LDAC catalog given, with vignets to create a PSF.

    cat_name : filename of the catalog
    output_cat : filename of the output catalog
    dir_chckimg : path to folder to save checkimages (will be created if non-existing)
    dir_psfex : path to folder to save the PSF (will be created if non-existing)
    config_file : filename of the PSFEx configuration file
    fwhm_range : range of FWHM (in pixels) used for PSFEx autoselect. 
                 Can be large if using a point-source only catalog
    min_snr : minimum SNR for PSFEx autoselect
    save_psf_png : save a png image of the PSF as checkimage
    verbose : verbose parameter
    """
    os.makedirs(dir_chckimg, exist_ok=True)
    os.makedirs(dir_psfex, exist_ok=True)
    study_name = ".".join(cat_name.split("/")[-1].split(".")[:-1])
    output_chk = f"{dir_chckimg}/{study_name}"
    verbose_type = 'NORMAL' if verbose else 'QUIET'
    subprocess.run(f'psfex {cat_name} -c {config_file} \
                   -CHECKIMAGE_NAME {dir_chckimg}/chi2.fits,{dir_chckimg}/samp.fits,{dir_chckimg}/res.fits,{dir_chckimg}/snap.fits \
                   -CHECKPLOT_NAME {output_chk}_selfwhm,{output_chk}_chi2,{output_chk}_counts \
                   -XML_NAME {output_chk}_xml.xml \
                   -OUTCAT_NAME {output_cat}_psf_cat.fits \
                   -PSF_DIR {dir_psfex} \
                   -SAMPLE_FWHMRANGE "{fwhm_range}" \
                   -SAMPLE_MINSN {min_snr} \
                   -VERBOSE_TYPE {verbose_type}',
                   shell=True)
    filename = f"{dir_psfex}/{study_name}_psf.psf"
    if save_psf_png:
        image = fits.open(filename)
        psf = image[1].data[0][0][0]
        image.close()
        fig, ax = plt.subplots(figsize=(6,6))
        ax.set_axis_off()
        ax.imshow(psf, origin='lower', cmap='bone_r', norm=SymLogNorm(linthresh=1e-5, linscale=0.8))
        fig.savefig(f"{filename.replace('.psf','.png')}", bbox_inches='tight', pad_inches=0, dpi=60)
        if verbose: plt.show()
    return filename
    
def compare_star(cat_name, cat_name_star, 
                 output_cat, output_cat_star, 
                 dir_chckimg, dir_psfex, config_file, 
                 fwhm_range='1.0,6.0', fwhm_range_star='1.0,10.0',
                 min_snr=200, min_snr_star=10,
                 verbose=False):
    """
    Function to compare the PSF between PSFEx standard auto-selection of point-like sources
    and a given point-like sources catalog.

    cat_name : filename of the raw SExtractor catalog
    cat_name_star : filename of the point-like sources catalog
    output_cat : filename of the output catalog
    output_cat_star : filename of the output point-like sources catalog
    dir_chckimg : path to folder to save checkimages (will be created if non-existing)
    dir_psfex : path to folder to save the PSF (will be created if non-existing)
    config_file : filename of the PSFEx configuration file
    fwhm_range : range of FWHM (in pixels) used for PSFEx autoselect
    fwhm_range_star : range of FWHM (in pixels) used for PSFEx autoselect (still used, with a large range)
    min_snr : minimum SNR for PSFEx autoselect
    min_snr_star : minimum SNR for PSFEx autoselect (still used, with a small value)
    verbose : verbose parameter
    """
    run_psfex(cat_name, output_cat, dir_chckimg, dir_psfex, config_file, fwhm_range=fwhm_range, min_snr=min_snr, verbose=verbose)
    run_psfex(cat_name_star, output_cat_star, dir_chckimg, dir_psfex, config_file, fwhm_range=fwhm_range_star, min_snr=min_snr_star, verbose=verbose)
    study_name = ".".join(cat_name.split("/")[-1].split(".")[:-1])
    study_name_star = ".".join(cat_name_star.split("/")[-1].split(".")[:-1])
    image = fits.open(f"{dir_psfex}/{study_name}_psf.psf")
    image_star = fits.open(f"{dir_psfex}/{study_name_star}_psf.psf")
    psf = np.array(image[1].data[0][0][0])
    psf_star = np.array(image_star[1].data[0][0][0])
    fig, ax = plt.subplots(1,2,figsize=(12,6))
    ax[0].imshow(psf, origin='lower', cmap='bone_r', norm=SymLogNorm(linthresh=1e-5, linscale=0.8))
    ax[1].imshow(psf_star, origin='lower', cmap='bone_r', norm=SymLogNorm(linthresh=1e-5, linscale=0.8))
    ax[0].set_title(f"PSFEx auto-selection \n SAMPLE_FWHMRANGE = ({fwhm_range})")
    ax[1].set_title("MU_MAX v. MAG_AUTO selection")
    plt.show()
    return fig

def main():
    pass

if __name__=="__main__":
    main()


import re
import os
import gzip
import shutil
import boto3

def find_files(bucket:str, path:str, regex:str) -> list[str]:
    """
    Find all files in a bucket matching a given regular expression.

    bucket : Bucket to search files in
    path : Path to search files in
    regex : Regular expression to match

    Returns : List of file names (without path)
    """
    files = []
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(bucket)
    for obj in bucket.objects.filter(Prefix=f"{path}/"):
        if re.fullmatch(regex, obj.key[len(path)+1:]):
            files.append(obj.key[len(path)+1:])
    return files

def decompress_save(file_name:str, 
                    in_bucket:str, in_path:str,
                    out_path:str,
                    verbose=False):
    """
    Downloads a compressed .gz file from an S3 bucket, 
    and saves a decompressed version to your computer.

    file_name : Name of the compressed file
    in_bucket : Bucket where the compressed file is
    in_path : Path in the in_bucket to the compressed file
    out_path : Path in your computer to save the decompressed file
    """
    s3 = boto3.client('s3')
    os.makedirs(out_path, exist_ok=True)
    # Download compressed images from S3
    if verbose: print(f"Downloading : https://s3.amazonaws.com/{in_bucket}/{in_path}/{file_name}")
    s3.download_file(in_bucket, f"{in_path}/{file_name}", f"{out_path}/{file_name}")
    # Decompress images
    if verbose: print(f"Decompressing : {out_path}/{file_name}")
    with gzip.open(f"{out_path}/{file_name}", 'rb') as image_gz:
        with open(f"{out_path}/{file_name[:-3]}", 'wb') as image_fits:
            shutil.copyfileobj(image_gz, image_fits)
    # Delete compressed images
    if verbose: print(f"Removing : {out_path}/{file_name}")
    os.remove(f"{out_path}/{file_name}")

def decompress_save_to_S3(file_name:str, 
                          in_bucket:str, in_path:str,
                          out_bucket:str, out_path:str,
                          temp_folder:str,
                          deleting_file=True,
                          verbose=False):
    """
    Downloads a compressed .gz file from an S3 bucket, 
    and saves a decompressed version in another S3 bucket.

    file_name : Name of the compressed file
    in_bucket : Bucket where the compressed file is
    in_path : Path in the in_bucket to the compressed file
    out_bucket : Bucket to save the decompressed file to
    out_path : Path in the out_bucket to save the decompressed file
    temp_folder : Temporary folder to manipulate files
    """
    s3 = boto3.client('s3')
    # Download and decompress images from S3
    decompress_save(file_name=file_name, 
                    in_bucket=in_bucket, in_path=in_path,
                    out_path=temp_folder, verbose=verbose)
    #Save decompressed images to S3
    if verbose: print(f"Uploading : https://s3.amazonaws.com/{out_bucket}/{out_path}/{file_name[:-3]}")
    s3.upload_file(f"{temp_folder}/{file_name[:-3]}", out_bucket, f"{out_path}/{file_name[:-3]}")
    # Delete decompressed images
    if deleting_file:
        if verbose: print(f"Removing : {temp_folder}/{file_name[:-3]}")
        os.remove(f"{temp_folder}/{file_name[:-3]}")


def save_s3(file, bucket, path):
    """
    Save a file to S3 bucket (with the same name)
    
    file : filename of the file to save
    bucket : S3 bucket name
    path : path in the S3 bucket to save the file"""
    s3 = boto3.client('s3')
    name = file.split("/")[-1]
    s3.upload_file(file, bucket, f"{path}/{name}")